/*
 * Copyright@ Samsung Electronics Co. LTD
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or distributed, transmitted,
 * transcribed, stored in a retrieval system or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed
 * to third parties without the express written permission of Samsung Electronics.


 * Alternatively, this program is free software in case of open source projec;
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.

*/

#include "bl2_config.h"
#include "board_config.h"
#include "bl2_common.h"
#include "clk-boot.h"
#include "cmusfr.h"
#include "bl2_smc.h"
#include "preset_qos.h"
#include "smc_aarch64.h"

#define AP_COLD_BOOT			0
#define AP_RESUME_MIF_RESET		1
#define CP_RESUME_MIF_RESET		2
#define RESUME_MIF_MUX_DONE		3

/* Clock function macro */
#define MDIV_SHIFT			(16)
#define PDIV_SHIFT			(8)
#define SDIV_SHIFT			(0)
#define KDIV_SHIFT			(0)
#define LOCKED_SHIFT			(29)
#define LOCK_VALUE(p)			(p * 150)
#define CON0_VALUE(m, p, s)		(((m) << MDIV_SHIFT) | ((p) << PDIV_SHIFT) | ((s) << SDIV_SHIFT) | (1 << 31))
#define WAIT_FOR_LOCKED(sfr)		while (!(__raw_readl(sfr) & (1 << LOCKED_SHIFT)))

#define MUX_BUSY			(26)
#define MUX_SET_SRC(src, sfr)		__raw_writel((1 << 21) | src << 12, sfr)
#define WAIT_FOR_MUX_SET_DONE(sfr)	while (__raw_readl(sfr) & (1 << MUX_BUSY))

#define DIV_BUSY			(25)
#define DIV_SET_VAL(val, sfr)		__raw_writel(val, sfr)
#define WAIT_FOR_DIV_SET_DONE(sfr)	while (__raw_readl(sfr) & (1 << DIV_BUSY))

#define ROOTCLKGATE_ENABLE
#define HWACG_ENABLE

#define PMU_ALIVE_BASE			0x11C80000
#define APM_MIF_ACCESS_WIN0		(PMU_ALIVE_BASE + 0x8610)
#define CORTEXM0_APM_OPTION		(PMU_ALIVE_BASE + 0x8108)

#define setbit(addr, bit, value)	__raw_writel((__raw_readl(addr) & ~(1 << bit)) | ((value & 0x1) << bit), addr)
#define setf(addr, bit, mask, value)	__raw_writel((__raw_readl(addr) & ~(mask << bit)) | ((value & mask) << bit), addr)
#define getf(addr, bit, mask)		((__raw_readl(addr) >> (bit)) & (mask))

typedef enum {
	DENSITY_4Gb = 0x6,
	DENSITY_8Gb, /* 0x7 */
	DENSITY_16Gb, /* 0x8 */
	DENSITY_32Gb, /* 0x9 */
	DENSITY_6Gb = 0xE,
} chip_density_t;

typedef enum {
	IO_32BIT = 0,
	IO_16BIT = 1,
} chip_io_t;

typedef enum {
	JAVA = 0,
	JAVA_W_SIP,
	JAVA_W,
} device_type_t;

void *memset8(void *dest, unsigned char data, int n)
{
	unsigned char *d = (unsigned char *)dest;

	while (n-- > 0) {
		*d++ = data;
	}

	return dest;
}

/*
CMU
0x0205_3000  ~ 0x0205_33FF   =>   0xD000 ~ 0xD3FF
*/

#define CMU_RESERVED	0x02053000

static struct cmu_resume_value *cmu_resume_value = (struct cmu_resume_value *)CMU_RESERVED;
//static struct cmu_cp_resume_value *cmu_cp_resume_value = (struct cmu_cp_resume_value *)(CMU_RESERVED + 0x50);
static unsigned char *device_type = (unsigned char *)(CMU_RESERVED + 0x60);

#define OTP_CON_TOP_BASE	(void *)(0x100D0000)

void delay(volatile unsigned int cnt)
{
	while (cnt--);
}

void cmu_cpucl0_init(unsigned int cmu_resume)
{
	unsigned int index;
	index = (cmu_resume == AP_COLD_BOOT) ? 0 : 1;

	MUX_SET_SRC(0, CLK_CON_MUX_CPUCL0_PLL);

	__raw_writel(LOCK_VALUE(cmu_resume_value->cpucl0_mps[(index * 3) + 1]), CPUCL0_PLL_LOCK);
	__raw_writel(CON0_VALUE(cmu_resume_value->cpucl0_mps[(index * 3) + 0],
			cmu_resume_value->cpucl0_mps[(index * 3) + 1],
			cmu_resume_value->cpucl0_mps[(index * 3) + 2]), CPUCL0_PLL_CON0);

	DIV_SET_VAL(0, CLK_CON_DIV_CLK_CPUCL0_1);
	DIV_SET_VAL(0, CLK_CON_DIV_CLK_CPUCL0_2);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_ACLK],
							CLK_CON_DIV_CLK_CPUCL0_ACLK);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_PCLK],
							CLK_CON_DIV_CLK_CPUCL0_PCLK);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_ATCLK],
							CLK_CON_DIV_CLK_CPUCL0_ATCLK);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_PCLKDBG],
							CLK_CON_DIV_CLK_CPUCL0_PCLKDBG);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_CNTCLK],
							CLK_CON_DIV_CLK_CPUCL0_CNTCLK);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_RUN_MONITOR],
							CLK_CON_DIV_CLK_CPUCL0_RUN_MONITOR);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_HPM],
							CLK_CON_DIV_CLK_CPUCL0_HPM);
	DIV_SET_VAL(cmu_resume_value->dvfs_cpucl0[(index * NUM_OF_CPUCL0_MEMBER) + CPUCL0_DIV_CLK_CPUCL0_PLL],
							CLK_CON_DIV_CLK_CPUCL0_PLL);

	WAIT_FOR_LOCKED(CPUCL0_PLL_CON0);

	MUX_SET_SRC(1, CLK_CON_MUX_CPUCL0_PLL);
	MUX_SET_SRC(0, CLK_CON_MUX_CLK_CPUCL0);
}

void cmu_g3d_init(unsigned int cmu_resume)
{
	unsigned int index;
	index = (cmu_resume == AP_COLD_BOOT) ? 0 : 1;

	DIV_SET_VAL(cmu_resume_value->dvfs_g3d[(index * NUM_OF_G3D_MEMBER) + G3D_DIV_CLK_G3D_BUS],
						CLK_CON_DIV_CLK_G3D_BUS);
	DIV_SET_VAL(cmu_resume_value->dvfs_g3d[(index * NUM_OF_G3D_MEMBER) + G3D_DIV_CLK_G3D_APB],
						CLK_CON_DIV_CLK_G3D_APB);

	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_G3D_USER);
}

void cmu_dispaud_init(unsigned int cmu_resume)
{
	/* AUD PLL 98.30402Mhz */
	__raw_writel(1 * 3000, AUD_PLL_LOCK);
	__raw_writel(CON0_VALUE(30, 1, 3), AUD_PLL_CON0);
	__raw_writel(16213, AUD_PLL_CON1);

#if defined(CONFIG_DECON_LCD_S6E8AA0)
	DIV_SET_VAL(3, CLK_CON_DIV_CLKCMU_DISPAUD_BUS);
	DIV_SET_VAL(13, CLK_CON_DIV_CLKCMU_DISPAUD_DECON_INT_VCLK);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_BUS);
	MUX_SET_SRC(2, CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK);

	DIV_SET_VAL(1, CLK_CON_DIV_CLK_DISPAUD_APB);
	DIV_SET_VAL(0, CLK_CON_DIV_CLK_DISPAUD_DECON_INT_VCLK);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_DISPAUD_BUS_USER);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK_USER);
#elif defined(CONFIG_DECON_LCD_AUO_H120BLN017)
	DIV_SET_VAL(3, CLK_CON_DIV_CLKCMU_DISPAUD_BUS);
	DIV_SET_VAL(0xd, CLK_CON_DIV_CLKCMU_DISPAUD_DECON_INT_VCLK);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_BUS);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK);

	DIV_SET_VAL(1, CLK_CON_DIV_CLK_DISPAUD_APB);
	DIV_SET_VAL(7, CLK_CON_DIV_CLK_DISPAUD_DECON_INT_VCLK);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_BUS_USER);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK_USER);
#elif defined(CONFIG_DECON_LCD_S6E8AA5X01)
	DIV_SET_VAL(3, CLK_CON_DIV_CLKCMU_DISPAUD_BUS);
	DIV_SET_VAL(3, CLK_CON_DIV_CLKCMU_DISPAUD_DECON_INT_VCLK);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_BUS);
	MUX_SET_SRC(0, CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK);

	DIV_SET_VAL(1, CLK_CON_DIV_CLK_DISPAUD_APB);
	DIV_SET_VAL(4, CLK_CON_DIV_CLK_DISPAUD_DECON_INT_VCLK);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_DISPAUD_BUS_USER);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK_USER);
#endif

	MUX_SET_SRC(1, CLK_CON_MUX_CLKPHY_DISPAUD_MIPIPHY_TXBYTECLKHS_USER);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKPHY_DISPAUD_MIPIPHY_RXCLKESC0_USER);

	DIV_SET_VAL(0, CLK_CON_DIV_CLK_DISPAUD_MI2S);
	DIV_SET_VAL(1, CLK_CON_DIV_CLK_DISPAUD_MIXER);

	WAIT_FOR_LOCKED(AUD_PLL_CON0);
	MUX_SET_SRC(1, CLK_CON_MUX_AUD_PLL);
}

void cmu_mfcmscl_init(unsigned int cmu_resume)
{
	DIV_SET_VAL(2, CLK_CON_DIV_CLK_MFCMSCL_APB);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_MFCMSCL_MSCL_USER);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_MFCMSCL_MFC_USER);
}

void cmu_isp_init(unsigned int cmu_resume)
{
	DIV_SET_VAL(1, CLK_CON_DIV_CLK_ISP_CAM_HALF);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_ISP_VRA_USER);
	MUX_SET_SRC(1, CLK_CON_MUX_CLKCMU_ISP_CAM_USER);
}

void cmu_fsys_init(unsigned int cmu_resume)
{
	/* USB_PLL = WPLL 20Mhz */
	if (cmu_resume == AP_COLD_BOOT) {
		/* change WPLL input selection to AP (AP2WLBT_USBPLL_WPLL_SEL). */
		setf(USBPLL_CON1, 0, 0x1, 0x1);

		if (getf(PMU_ALIVE_BASE + 0x7824, 8, 0x1) == 1) {
		/* TCXO_SHARED_STATUS = 52Mhz */
			setf(USBPLL_CON1,3,0x1,0x1);
		} else {
		/* 26Mhz */
			setf(USBPLL_CON1,3,0x1,0x0);
		}

		/* Set WPLL enable (AP2WLBT_USBPLL_WPLL_EN) */
		setf(USBPLL_CON1,1,0x1,0x1);

		/* wait 20us for power settle time. */
		/* need mct delay function */
		delay(20000);

		/* Set WPLL AFC Start (AP2WLBT_USBPLL_WPLL_AFC_START) */
		setf(USBPLL_CON1,2,0x1,0x1);

		/* wait 60us for clock stabilization. */
		/* need mct delay function */
		delay(60000);
	}
}

void mux_enable_loop(void *start, int loop)
{
	int i, offset;

	for (i = 0, offset = 0; i <= loop; i++, offset += 4)
		setbit(start + offset, 21, 0x1);
}

void top_mux_gating_disable(unsigned int cmu_resume)
{
	if (cmu_resume != AP_COLD_BOOT) {
		mux_enable_loop(CLK_CON_MUX_CLKCMU_G3D, 4);
		mux_enable_loop(CLK_CON_MUX_CLKCMU_MFCMSCL_MSCL, 6);
		mux_enable_loop(CLK_CON_MUX_CLKCMU_PERI_UART_DEBUG, 1);
		mux_enable_loop(CLK_CON_MUX_CLKCMU_PERI_SPI_REARFROM, 5);
/*
		setbit(CLK_CON_MUX_CLKCMU_G3D, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_ISP_VRA, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_ISP_CAM, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_DISPAUD_BUS, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_DISPAUD_DECON_INT_VCLK, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_MFCMSCL_MFC, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_FSYS_BUS, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_FSYS_MMC0, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_FSYS_MMC2, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_FSYS_USB20DRD_REFCLK, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_BUS, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_UART_DEBUG, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_UART_SENSOR, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_SPI_REARFROM, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_SPI_ESE, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_USI_0, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_PERI_USI_1, 21, 0x1);
		setbit(CLK_CON_MUX_CLKCMU_ISP_SENSOR0, 21, 0x1);
*/
	}
}

int clk_init(unsigned int cmu_resume)
{
	if (cmu_resume != RESUME_MIF_MUX_DONE)
		cmu_cpucl0_init(cmu_resume);

	if (cmu_resume == AP_COLD_BOOT) {
		cmu_g3d_init(cmu_resume);
		cmu_dispaud_init(cmu_resume);
		cmu_mfcmscl_init(cmu_resume);
		cmu_isp_init(cmu_resume);
		cmu_fsys_init(cmu_resume);
	}

	top_mux_gating_disable(cmu_resume);

	return 0;
}

void chipid_get_device_type(void)
{
/*	unsigned int uBits;
	unsigned char type;
	uBits = (unsigned int)exynos_smc_read(SMC_CMD_REG, SMC_REG_ID_SFR_R((unsigned long long)(OTP_CON_TOP_BASE + 0xB)));
	type = (unsigned char)((uBits >> 5) & 0x1F);
	*device_type = type;
	enable when bit fused */
#if (defined (CONFIG_MACH_ESPRESSO7270_SIP) || defined (CONFIG_MACH_UNIVERSAL7270_SIP))
	*device_type = JAVA_W_SIP;
#elif (defined (CONFIG_MACH_ESPRESSO7270) || defined (CONFIG_MACH_UNIVERSAL7270))
	*device_type = JAVA_W;
#elif (defined (CONFIG_MACH_ESPRESSO7570) || defined (CONFIG_MACH_UNIVERSAL7570))
	*device_type = JAVA;
#endif
}

void apm_cmu_dmc_init(void)
{
	unsigned int apm_mif_access_win0;

	/* APM WDT disabled */
	__raw_writel(0x0, 0x11C60000);
	/* CONFIG_APM, REMAP0 Start address of code region */
	__raw_writel(0x02046000, 0x11c50000);
	/* CONFIG_APM, REMAP2 Size of code/data 14*4KB */
	__raw_writel(0xe, 0x11c50008);

	/* [13:12]=CMU_MIF [11:10]=SYSREG_MIF [9:8]=DDRPHY0 [3:2]=DREX0_PF [1:0}=DREX0 - R/W enable */
	apm_mif_access_win0 = __raw_readl(APM_MIF_ACCESS_WIN0);
	apm_mif_access_win0 |= ((3<<12) | (3<<10) |(3<<8) | (3<<2) | (3<<0));
	__raw_writel(apm_mif_access_win0, APM_MIF_ACCESS_WIN0);

	memset8(cmu_resume_value, 0x0, sizeof(struct cmu_resume_value));
/*	memset8(cmu_cp_resume_value, 0x0, sizeof(struct cmu_cp_resume_value));*/

	__raw_writel(0x0, 0x11C00080);
	/* APM enable */
	__raw_writel(1<<15, CORTEXM0_APM_OPTION);
	/* MAILBOX_APM2AP : SR0 */
	while (__raw_readl(0x11C00080) != 0x1234ABCD);
	__raw_writel(0x1, 0x11C00080);
}

void drex_config_by_dram_info(void) {
	reg_arr_t reg_arr;
	unsigned int DREX_CALIBRATION0;
	unsigned int cs0_mem_size, cs1_mem_size;
	unsigned int cs0_io_width, cs1_io_width;
	unsigned int ADDR_MAP_CONFIG_EN, CHUNK1_EN, CHUNK0_START, CHUNK0_SIZE;	/* EXT_MEMCFG0 */
	unsigned int RANK0_SIZE, RANK0_ROW_BW, RANK0_COL_BW, RANK_INTER_EN = 0;	/* EXT_MEMCFG1 */
	unsigned int RANK1_ROW_BW, RANK1_COL_BW, RANK1_EN = 0; /* EXT_MEMCFG2 */
	unsigned int uRANK1_SIZE;

	DREX_CALIBRATION0 = __raw_readl(0x11c809a0);
	cs0_mem_size = ((DREX_CALIBRATION0 & (0xF << 20)) >> 20);
	cs0_io_width = ((DREX_CALIBRATION0 & (0x3 << 28)) >> 28);
	cs1_mem_size = ((DREX_CALIBRATION0 & (0xF << 24)) >> 24);
	cs1_io_width = ((DREX_CALIBRATION0 & (0x3 << 30)) >> 30);

	/* rank inter en */
	if (cs0_mem_size == cs1_mem_size) {
		if (cs0_io_width == cs1_io_width) {
			RANK_INTER_EN = 1;
		}
	}

	switch (cs0_mem_size) {
	case DENSITY_4Gb:
		RANK0_ROW_BW = 0;
		if(cs0_io_width == IO_32BIT) {
			RANK0_COL_BW = 0;
			RANK0_SIZE = 4; /* 3; //0xn: (n+1)*128MiB */
		} else {
			RANK0_COL_BW = 1;
			RANK0_SIZE = 8; /* 7; //0xn: (n+1)*128MiB */
		}
		break;
	case DENSITY_6Gb:
		RANK0_ROW_BW = 1;
		if(cs0_io_width == IO_32BIT) {
			RANK0_COL_BW = 0;
			RANK0_SIZE = 6; /* 5; //0xn: (n+1)*128MiB */
		} else {
			RANK0_COL_BW = 1;
			RANK0_SIZE = 12; /* 11; //0xn: (n+1)*128MiB */
		}
		break;
	case DENSITY_8Gb:
		RANK0_ROW_BW = 1;
		if(cs0_io_width == IO_32BIT) {
			RANK0_COL_BW = 0;
			RANK0_SIZE = 8; /* 7; //0xn: (n+1)*128MiB */
		} else {
			RANK0_COL_BW = 1;
			RANK0_SIZE = 16; /* 15; //0xn: (n+1)*128MiB */
		}
		break;
	default: /* no rank detected */
		RANK0_ROW_BW = 0;
		RANK0_COL_BW = 0;
		RANK0_SIZE = 0;
		CHUNK1_EN = 0;
		break;
	}

	switch (cs1_mem_size) {
	case DENSITY_4Gb:
		RANK1_ROW_BW = 0;
		if(cs1_io_width == IO_32BIT) {
			RANK1_COL_BW = 0;
			uRANK1_SIZE = 4; /* 3; */
		} else {
			RANK1_COL_BW = 1;
			uRANK1_SIZE = 8; /* 7; */
		}
		break;
	case DENSITY_6Gb:
		RANK1_ROW_BW = 1;
		if(cs1_io_width == IO_32BIT) {
			RANK1_COL_BW = 0;
			uRANK1_SIZE = 6; /* 5; */
		} else {
			RANK1_COL_BW = 1;
			uRANK1_SIZE = 12; /* 11; */
		}
		break;
	case DENSITY_8Gb:
		RANK1_ROW_BW = 1;
		if(cs1_io_width == IO_32BIT) {
			RANK1_COL_BW = 0;
			uRANK1_SIZE = 8; /*7; */
		} else {
			RANK1_COL_BW = 1;
			uRANK1_SIZE = 16; /* 15; */
		}
		break;
	default: /* no rank detected */
		RANK1_ROW_BW = 0;
		RANK1_COL_BW = 0;
		uRANK1_SIZE = 0;
		CHUNK1_EN = 0;
		break;
	}

	if(cs1_mem_size !=0) {
		RANK1_EN =1 ;
	}

	ADDR_MAP_CONFIG_EN = 1; //1'b1: Use configurations from external pin
	CHUNK0_START = 8; //The start address[31:27] of chunk 0
	CHUNK0_SIZE = (RANK0_SIZE + uRANK1_SIZE - 1); //(RANK0_SIZE + uRANK1_SIZE +1);
	CHUNK1_EN = 0;
	if (RANK0_SIZE > 0)
		RANK0_SIZE = RANK0_SIZE - 1;

	reg_arr.set0.addr = 0x11C87100; /* EXT_MEMCFG0 */
	reg_arr.set0.val = ((ADDR_MAP_CONFIG_EN << 31) | (CHUNK1_EN << 30) | (CHUNK0_START << 24) | (CHUNK0_SIZE << 16));
	reg_arr.set1.addr = 0x11C87104; /* EXT_MEMCFG1 */
	reg_arr.set1.val = ((RANK_INTER_EN << 24) | (RANK0_COL_BW << 8) | (RANK0_ROW_BW << 4) | (RANK0_SIZE << 0));
	reg_arr.set2.addr = 0x11C87108; /* EXT_MEMCFG2 */
	reg_arr.set2.val = ((RANK1_EN << 31) | (RANK1_COL_BW << 8) | (RANK1_ROW_BW << 4));
	set_secure_reg((u64)&reg_arr, 3);
}

void dram_size_config_by_dram_info(void) {
	unsigned int DREX_CALIBRATION0;
	unsigned int cs0_mem_size, cs1_mem_size;
	unsigned int cs0_io_width, cs1_io_width;
	unsigned int uCS0_SIZE, uCS1_SIZE;

	DREX_CALIBRATION0 = __raw_readl(0x11c809a0);
	cs0_mem_size = ((DREX_CALIBRATION0 & (0xF << 20)) >> 20);
	cs0_io_width = ((DREX_CALIBRATION0 & (0x3 << 28)) >> 28);
	cs1_mem_size = ((DREX_CALIBRATION0 & (0xF << 24)) >> 24);
	cs1_io_width = ((DREX_CALIBRATION0 & (0x3 << 30)) >> 30);

	switch (cs0_mem_size) {
	case DENSITY_4Gb:
		if (cs0_io_width == IO_32BIT) {
			uCS0_SIZE = 0x20000000;	/* 512MB */
		} else if (cs0_io_width == IO_16BIT) {
			uCS0_SIZE = 0x40000000;	/* 512MB*2 */
		} else {
			uCS0_SIZE = 0;
		}
		break;
	case DENSITY_6Gb:
		if (cs0_io_width == IO_32BIT) {
			uCS0_SIZE = 0x30000000;	/* 768MB */
		} else if (cs0_io_width == IO_16BIT) {
			uCS0_SIZE = 0x60000000;	/* 768MB*2 */
		} else {
			uCS0_SIZE = 0;
		}
		break;
	case DENSITY_8Gb:
		if (cs0_io_width == IO_32BIT) {
			uCS0_SIZE = 0x40000000;	/* 1024MB */
		} else if (cs0_io_width == IO_16BIT) {
			uCS0_SIZE = 0x80000000;	/* 1024MB*2 */
		} else {
			uCS0_SIZE = 0;
		}
		break;
	default: /* no rank detected */
		uCS0_SIZE = 0;
		break;
	}

	switch (cs1_mem_size) {
	case DENSITY_4Gb:
		if (cs1_io_width == IO_32BIT) {
			uCS1_SIZE = 0x20000000;	/* 512MB */
		} else if (cs1_io_width == IO_16BIT) {
			uCS1_SIZE = 0x40000000;	/* 512MB*2 */
		} else {
			uCS1_SIZE = 0;
		}
		break;
	case DENSITY_6Gb:
		if (cs1_io_width == IO_32BIT) {
			uCS1_SIZE = 0x30000000;	/* 768MB */
		} else if (cs1_io_width == IO_16BIT) {
			uCS1_SIZE = 0x60000000;	/* 768MB*2 */
		} else {
			uCS1_SIZE = 0;
		}
		break;
	case DENSITY_8Gb:
		if (cs1_io_width == IO_32BIT) {
			uCS1_SIZE = 0x40000000;	/* 1024MB */
		} else if (cs1_io_width == IO_16BIT) {
			uCS1_SIZE = 0x80000000;	/* 1024MB*2 */
		} else {
			uCS1_SIZE = 0;
		}
		break;
	default: /* no rank detected */
		uCS1_SIZE = 0;
		break;
	}

	*((volatile unsigned long long *)CONFIG_PHY_BL_SYS_INFO_DRAM_SIZE) = (unsigned long long)(uCS0_SIZE + uCS1_SIZE);
}

void cmu_dmc_init(void)
{
	unsigned int uReg;

	if (__raw_readl(0x11C80600) == 0) { /* read wakeup stat */
		chipid_get_device_type();
		apm_cmu_dmc_init();
		clk_init(AP_COLD_BOOT);
		/* Memory size */
		drex_config_by_dram_info();
		dram_size_config_by_dram_info();

		uReg = __raw_readl(0x10400010);	/* MEMORY_CONTROL.ref_rate_con_en */
		uReg |= (1 << 20);
		__raw_writel(uReg, 0x10400010);

		uReg = __raw_readl(0x10400008);	/* CONTROLLER_CONTROL.aref_en */
		uReg |= (1 << 1);
		__raw_writel(uReg, 0x10400008);

		__raw_writel(0x1, 0x10400420); /* ALL_INIT_INDICATION all_init_done */
		__raw_writel(0x1, 0x10410004); /* PF_ALL_INIT_INDICATION all_init_done */
		__raw_writel(0x1, 0x11C809AC); /* DREX_CALIBRATION3 */
		decon_bts_conf();
	} else if (getf(0x11C8112C, 0, 0x3) == 0) {
		/* RESET_CMU_MIF_SYS_PWR_REG [1:0] == 0 All cmu does not retain setting  */
		clk_init(AP_RESUME_MIF_RESET);
		/* Wake up from DSTOP-MOFF, SLEEP, inform Memory size */
		dram_size_config_by_dram_info();
		__raw_writel(0x1, 0x11C00080);
	} else if ( __raw_readl(0x10400420) == 0x1) {
		clk_init(RESUME_MIF_MUX_DONE);
	}
}

