/*
 * Driver for Samsung SPEEDY(Serial Protocol in an EffEctive Digital waY)
 *
 * Copyright@ Samsung Electronics Co. LTD
 *      Youngmin Nam <youngmin.nam@samsung.com>
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or
 * distributed, transmitted, transcribed, stored in a retrieval system or
 * translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed to third parties
 * without the express written permission of Samsung Electronics.

 * Alternatively, this program is free software in case of open source project;
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "bl2_io.h"
#include "speedy.h"

static void delay(volatile unsigned int cnt)
{
	while (cnt--);
}

static void speedy_set_enable(void)
{
	unsigned int speedy_ctl = readl(SPEEDY_BASE + SPEEDY_CTRL);

	speedy_ctl |= SPEEDY_ENABLE;
	writel(speedy_ctl, SPEEDY_BASE + SPEEDY_CTRL);
}

static void speedy_fifo_reset(void)
{
	unsigned int speedy_fifo_ctl = readl(SPEEDY_BASE + SPEEDY_FIFO_CTRL);

	speedy_fifo_ctl |= SPEEDY_FIFO_RESET;
	writel(speedy_fifo_ctl, SPEEDY_BASE + SPEEDY_FIFO_CTRL);
	/* 10us delay for speedy fifo reset */
	delay(10000);
}

static void speedy_set_cmd(int direction, unsigned short address, int random, int burst_length)
{
	unsigned int speedy_fifo_ctl = 0;
	unsigned int speedy_int_en = 0;
	unsigned int speedy_command = 0;

	speedy_fifo_reset();
	speedy_fifo_ctl = readl(SPEEDY_BASE + SPEEDY_FIFO_CTRL);

	speedy_command |= SPEEDY_ADDRESS(address);

	switch (random) {
	case ACCESS_RANDOM:
		speedy_command |= SPEEDY_ACCESS_RANDOM;
		speedy_fifo_ctl |= (SPEEDY_RX_TRIGGER_LEVEL(1) |
				SPEEDY_TX_TRIGGER_LEVEL(1));
		break;
	case ACCESS_BURST:
		speedy_command |= (SPEEDY_ACCESS_BURST | SPEEDY_BURST_INCR |
				   SPEEDY_BURST_LENGTH(burst_length-1));
		speedy_fifo_ctl |= (
			SPEEDY_RX_TRIGGER_LEVEL(burst_length) |
			SPEEDY_TX_TRIGGER_LEVEL(1)
		);
		break;

	default:
		while(1);
		break;
	}

	/* configure speedy fifio trigger level */
	writel(speedy_fifo_ctl, SPEEDY_BASE + SPEEDY_FIFO_CTRL);

	speedy_int_en |= (SPEEDY_TIMEOUT_CMD_EN | SPEEDY_TIMEOUT_STANDBY_EN |
			SPEEDY_TIMEOUT_DATA_EN);

	switch (direction) {
	case DIRECTION_READ:
		speedy_command |= SPEEDY_DIRECTION_READ;
		speedy_int_en |= (SPEEDY_FIFO_RX_ALMOST_FULL_EN |
				SPEEDY_RX_FIFO_INT_TRAILER_EN |
				SPEEDY_RX_MODEBIT_ERR_EN |
				SPEEDY_RX_GLITCH_ERR_EN |
				SPEEDY_RX_ENDBIT_ERR_EN |
				SPEEDY_REMOTE_RESET_REQ_EN);
		break;

	case DIRECTION_WRITE:
		speedy_command |= SPEEDY_DIRECTION_WRITE;
		speedy_int_en |= (SPEEDY_TRANSFER_DONE_EN |
				SPEEDY_FIFO_TX_ALMOST_EMPTY_EN |
				SPEEDY_TX_LINE_BUSY_ERR_EN |
				SPEEDY_TX_STOPBIT_ERR_EN |
				SPEEDY_REMOTE_RESET_REQ_EN);
		break;

	default:
		while(1);
		break;
	}

	/* clear speedy interrupt status */
	writel(0xFFFFFFFF, SPEEDY_BASE + SPEEDY_INT_STATUS);

	/* speedy interrupt enable */
	writel(speedy_int_en, SPEEDY_BASE + SPEEDY_INT_ENABLE);

	/* configure speedy command */
	writel(speedy_command, SPEEDY_BASE + SPEEDY_CMD);
}

static int speedy_wait_transfer_done(void)
{
	unsigned int speedy_int_status;
	unsigned long timeout = 1000;
	int ret = 0;

	while (timeout-- > 0) {
		speedy_int_status = readl(SPEEDY_BASE + SPEEDY_INT_STATUS);

		if (speedy_int_status & SPEEDY_TRANSFER_DONE) {
			ret = 0;
			break;
		}
	}

	if (timeout < 0)
		ret = -1;

	writel(0xFFFFFFFF, SPEEDY_BASE + SPEEDY_INT_STATUS);

	return ret;
}

static void speedy_swreset(void)
{
	unsigned int speedy_ctl = readl(SPEEDY_BASE + SPEEDY_CTRL);

	speedy_ctl |= SPEEDY_SW_RST;
	writel(speedy_ctl, SPEEDY_BASE + SPEEDY_CTRL);
	/* 10us delay for speedy sw_rst */
	delay(10000);
}

void speedy_init(void)
{
	/* clear speedy interrupt status */
	writel(0xFFFFFFFF, SPEEDY_BASE + SPEEDY_INT_STATUS);

	/* reset speedy ctrl SFR. It may be used by bootloader */
	speedy_swreset();
	speedy_set_enable();
}

void speedy_read(unsigned char slave, unsigned char addr, unsigned char *data)
{
	int result;
	unsigned short address = addr + ((slave & 0xF) << 8);

	speedy_set_cmd(DIRECTION_READ, address, ACCESS_RANDOM, 1);

	result = speedy_wait_transfer_done();

	if (result == 0)
		*data = (unsigned char)readl(SPEEDY_BASE + SPEEDY_RX_DATA);
	else
		while(1);
}

void speedy_write(unsigned char slave, unsigned char addr, unsigned char data)
{
	int result;
	unsigned short address = addr + ((slave & 0xF) << 8);

	speedy_set_cmd(DIRECTION_WRITE, address, ACCESS_RANDOM, 1);
	writel(data, SPEEDY_BASE + SPEEDY_TX_DATA);

	result = speedy_wait_transfer_done();

	if (result == 0)
		return;
	else
		while(1);
}
