#include <asm/types.h>
#include <asm/arch/pmic_s2mpw01.h>
#include "../../../drivers/speedy/speedy.h"

typedef enum {false, true} bool;

enum {
	POWER_SUPPLY_CHARGE_TYPE_UNKNOWN = 0,
	POWER_SUPPLY_CHARGE_TYPE_NONE,
	POWER_SUPPLY_CHARGE_TYPE_TRICKLE,
	POWER_SUPPLY_CHARGE_TYPE_FAST,
	POWER_SUPPLY_CHARGE_TYPE_FULL,
};

#define CHARGE_SOURCE_USB       0
#define CHARGE_SOURCE_AC        1

struct s2mpw01_regs {
	u8 pmicid;
	u8 pmicver;
	u8 status1;
	u8 status2;
	u8 pwronsrc;
	u8 offsrc;
	u8 int1;
	u8 int2;
	u8 int3;
};

static struct s2mpw01_regs pmic_regs;


static u8 chip_revision;

static void s2mpw01_test_read(void)
{
	u8 reg_data;
	u8 addr;

	/* check register setting for debugging */

	printf("%s: ", __func__);

	for (addr = 0x03; addr <= 0x11; addr++) {
		PMIC_CHARGER_READ(addr, &reg_data);
		printf("0x%02x:0x%02x, ", addr, reg_data);
	}
	printf("\n");
}

int get_charger_state(void)
{
	int state = POWER_SUPPLY_CHARGE_TYPE_UNKNOWN;
	u8 chg_sts;

	/* get charging type */
	/* the state for charging type should be one of below
		enum {
		POWER_SUPPLY_CHARGE_TYPE_UNKNOWN = 0,
		POWER_SUPPLY_CHARGE_TYPE_NONE,
		POWER_SUPPLY_CHARGE_TYPE_TRICKLE,
		POWER_SUPPLY_CHARGE_TYPE_FAST,
		POWER_SUPPLY_CHARGE_TYPE_FULL,
	};*/

	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS1, &chg_sts);

	if (chg_sts & TOPOFF_STATUS_MASK)
		state = POWER_SUPPLY_CHARGE_TYPE_FULL;
	else if (chg_sts & TOPOFF_STATUS_MASK)
		state = POWER_SUPPLY_CHARGE_TYPE_FAST;
	else if (chg_sts & PRECHG_STATUS_MASK)
		state = POWER_SUPPLY_CHARGE_TYPE_TRICKLE;
	else if (!(chg_sts & CHGSTS_STATUS_MASK))
		state = POWER_SUPPLY_CHARGE_TYPE_NONE;

	printf("state = (%d)\n", state);

	return state;
}

int get_battery_detect(void)
{
	u8 data;
	/* check battery detect */
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS2, &data);

	return (data & DET_BAT_STATUS_MASK) ? true : false;
}

int get_vbus_state(void)
{
	u8 data;
	/* get vbus state. if vbus is valid, return 1 */
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS1, &data);

	return (data & CHGVIN_STATUS_MASK) ? true : false;
}

void set_regulation_voltage(int float_voltage)
{
	int data;
	u8 temp;

	if (float_voltage <= 4200)
		data = 0;
	else if (float_voltage > 4200 && float_voltage <= 4550)
		data = (float_voltage - 4200) / 50;
	else
		data = 0x7;

	printf("%s  float voltage : %d\n", __func__, float_voltage);

	PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL5, &temp);
	temp &= ~0x0E;
	temp |= (data << 3);
	PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL5, temp);
}

void set_charging_current(int charging_current)
{
	int data;
	int current;
	u8 temp;

	if (charging_current > 200)
		current = 200;

	if (current <= 150)
		data = 0;
	else if (current > 150 && current <= 400)
		data = (current - 150) / 50;
	else
		data = 0x5;

	printf("%s  charging_current: %d\n", __func__, charging_current);

	PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL2, &temp);
	temp &= ~0x0E;
	temp |= (data << 1);
	PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL2, temp);

	/* Additional Charging path on when Very low battery condtion */
	if (charging_current > 200) {
		/* Additional Current 7/8 * Main */
		temp = 0x7f;
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL7, temp);

		/* Additional Current Path On */
		PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL5, &temp);
		temp |= (1 << 5);
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL5, temp);
	} else {
		/* Additional Current Path Off */
		PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL5, &temp);
		temp &= ~(1 << 5);
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL5, temp);

		/* Additional Current 8/8 * Main */
		temp = 0xff;
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL7, temp);
	}
}

void set_topoff_current(int current_limit)
{
	int data;
	u8 temp;

	if (current_limit <= 5)
		data = 0;
	else if (current_limit > 5 && current_limit <= 10)
		data = (current_limit - 5) / 5;
	else if (current_limit > 10 && current_limit < 18)
		data = (current_limit - 10) / 5 * 2 + 1;
	else if (current_limit >= 18 && current_limit < 20)
		data = 5;	  /* 17.5 mA */
	else if (current_limit >= 20 && current_limit < 25)
		data = 4;
	else if (current_limit >= 25 && current_limit <= 40)
		data = (current_limit - 25) / 5 + 6;
	else if (current_limit > 40 && current_limit <= 100)
		data = (current_limit - 40) / 10 + 9;
	else
		data = 0x0F;

	printf("%s top-off current: %d\n", __func__, current_limit);

	PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL4, &temp);
	temp &= ~0xF0;
	temp |= (data << 4);
	PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL4, temp);
}

void set_charging_tmrout(int enable)
{
	u8 temp;

	if (enable) {
		PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL1, &temp);
		temp &= ~0x1C;
		temp |= 0x0C;
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL1, temp);
		printf("%s enabled 10 hours[0x%x]\n", __func__, temp);
	} else {
		PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL1, &temp);
		temp &= ~0x1C;
		temp |= 0x1C;
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL1, temp);
		printf("%s disabled[0x%x]\n", __func__, temp);
	}
}

/* W/A function to reload otp value
 * After this function, FG(0x26~0x2D) and CHG(0x17~0x34) values are reset.
 */
void s2mpw01_otp_reload(void)
{
	u8 temp;

	PMIC_COMMON_READ(0x10, &temp);
	temp |= 0x20;
	PMIC_COMMON_WRITE(0x10, temp);

	PMIC_COMMON_READ(0x0E, &temp);
	temp |= 0x80;
	PMIC_COMMON_WRITE(0x0E, temp);

	mdelay(100);

	PMIC_COMMON_READ(0x10, &temp);
	temp &= ~0x20;
	PMIC_COMMON_WRITE(0x10, temp);

	PMIC_COMMON_READ(0x0E, &temp);
	temp &= ~0x80;
	PMIC_COMMON_WRITE(0x0E, temp);

	printf("%s\n", __func__);
}

u8 pmic_get_pmic_chip_version(void)
{
	/* read pmic id */
	PMIC_COMMON_READ(S2MPW01_COMMON_PMIC_ID, &pmic_regs.pmicid);
	pmic_regs.pmicver = (pmic_regs.pmicid & S2MPW01_COMMON_PMIC_VERSION_MASK);
	printf("%s, PMIC CHIP Version : 0x%X \n", __func__, pmic_regs.pmicver);
	/* get the pmic chip version */
	return pmic_regs.pmicver;
}

void init_charger(void)
{
	u8 temp, data;
	u8 chg_stat2, chg_stat3;
	int factory_mode = 0;
	/* charger initialization */
	/* if don't need any special setting for init, just check i2c read like device id etc. */

	printf("%s: Run otp_reload\n", __func__);
	s2mpw01_otp_reload();

	PMIC_CHARGER_WRITE(0x25, 0x82);

	/* initialize the chip revision */
	chip_revision = pmic_get_pmic_chip_version();

	/* Increase debounce time & CHGIN to VBATT minimum turn on threshold level decrease */
	PMIC_CHARGER_WRITE(0x2B, 0x72);

	/* IVR Level decrease */
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL4, &temp);
	temp &= 0xF1;
	temp |= 0x00;
	PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL4, temp);

	if (chip_revision < EVT_3)
		PMIC_CHARGER_WRITE(0x28, 0x6F);

	/* MRSTB enable & WTSR enable */
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL9, &temp);
	temp &= 0x77;
	temp |= 0x88;
	PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL9, temp);

	set_charging_tmrout(1);

	/* factory_mode setting */
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS2, &chg_stat2);
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS3, &chg_stat3);
	printf("%s  factory mode : 0x%x, 0x%x\n", __func__, chg_stat2, chg_stat3);

	if ((chg_stat2 & 0x04) || (chg_stat2 & 0x02) ||
		(chg_stat2 & 0x01) || (chg_stat3 & 0x08)) {
		factory_mode = 1;
	}

	/* RID operation setting */
	PMIC_CHARGER_WRITE(0x20, 0x8A);

	/* ACOK status */
	if (chip_revision < EVT_3) {
		PMIC_READ(0x41, &data);
		if (data != 0)
			data &= 0xBF;
		else
			data = 0x21;

		PMIC_READ(S2MPW01_STATUS1, &temp);
		if ((temp & S2MPW01_PWRON_ACOK) && (factory_mode == 0)) {
			/* ACOK status High */
			data |= 0x40;
		}
		PMIC_CHARGER_WRITE(0x2E, data);
	}

	/* device_id : from Commom block 0x00[7:0] */
}

void set_charger_current(int type)
{
	u8 temp;
	/* defulat setting for charger */

	printf("%s\n", __func__);

	/* set 4.35V for float voltage */
	set_regulation_voltage(4350);

	/* set 50mA for topoff current */
	set_topoff_current(50);

	switch (type) {
	case CHARGE_SOURCE_AC:
	case CHARGE_SOURCE_USB:
	default:
		/* set 400mA for charging current */
		set_charging_current(400);
		break;
	}
}

void set_charger_state(int buck, int enable)
{
	u8 temp;
	u8 data;
	u8 chg_stat2, chg_stat3;
	int factory_mode = 0;

	/* factory_mode setting */
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS2, &chg_stat2);
	PMIC_CHARGER_READ(S2MPW01_CHG_REG_STATUS3, &chg_stat3);
	printf("%s  factory mode : 0x%x, 0x%x\n", __func__, chg_stat2, chg_stat3);

	if ((chg_stat2 & 0x04) || (chg_stat2 & 0x02) ||
		(chg_stat2 & 0x01) || (chg_stat3 & 0x08)) {
		factory_mode = 1;
	}

	if (chip_revision < EVT_3) {
		PMIC_READ(0x41, &data);
		if (data != 0)
			data &= 0xBF;
		else
			data = 0x21;
	}

	/* if enable == 1, set default charging current and enable charging */
	/* if enable == 0, disable charging */
	if (enable) {
		/* set 400mA for charging current */
		set_charging_current(400);

		/* charger enable */
		PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL1, &temp);
		temp |= 0x80;
		PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL1, temp);
	} else {
		if (chip_revision < EVT_3) {
			if (factory_mode == 0) {
				data |= 0x40;
				PMIC_CHARGER_WRITE(0x2E, data);

				/* charger disable */
				PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL1, &temp);
				temp &= ~0x80;
				PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL1, temp);
			}
		}

		/* ACOK status */
		if (chip_revision < EVT_3) {
			PMIC_READ(S2MPW01_STATUS1, &temp);
			if (!(temp & S2MPW01_PWRON_ACOK)) {
				/* ACOK status Low */

					data &= 0xBF;
					PMIC_CHARGER_WRITE(0x2E, data);

				/* charger enable */
				PMIC_CHARGER_READ(S2MPW01_CHG_REG_CTRL1, &temp);
				temp |= 0x80;
				PMIC_CHARGER_WRITE(S2MPW01_CHG_REG_CTRL1, temp);
			}
		}
	}
}
