/* drivers/video/exynos/7880/decon_core.c
 *
 * SAMSUNG SoC DECON driver for EXYNOS7880.
 *
 * Copyright@ Samsung Electronics Co. LTD
 *
 * Manseok Kim <manseoks.kim@samsung.com> *
 * Seuni Park <seuni.park@samsung.com> *
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or distributed, transmitted,
 * transcribed, stored in a retrieval system or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed
 * to third parties without the express written permission of Samsung Electronics.


 * Alternatively, this program is free software in case of open source projec;
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.

*/


#include <malloc.h>
#include "../decon.h"
#include "../dsim.h"
#include "decon_config.h"
#include "../panels/lcd_ctrl.h"
#ifdef CONFIG_DISPLAY_LOGO
#include "../decon_logo.h"
#endif

struct decon_device *decon_int_drvdata;
EXPORT_SYMBOL(decon_int_drvdata);
struct decon_device *decon_ext_drvdata;
EXPORT_SYMBOL(decon_ext_drvdata);

extern struct dsim_device *dsim0_for_decon;
extern struct dsim_device *dsim1_for_decon;

#ifdef CONFIG_DISPLAY_DRAWFONT
static struct decon_device *g_decon;
static struct dsim_device *g_dsim;
unsigned int *win_fb0 = CONFIG_DISPLAY_LOGO_BASE_ADDRESS;
#endif

static void decon_to_init_param(struct decon_device *decon, struct decon_param *p)
{
	struct decon_lcd *lcd_info = decon_get_lcd_info();

	p->lcd_info = lcd_info;
	p->psr.psr_mode = decon->pdata->psr_mode;
	p->psr.trig_mode = decon->pdata->trig_mode;
	p->psr.dsi_mode = decon->pdata->dsi_mode;
	p->nr_windows = decon->pdata->max_win;

	p->psr.out_type = decon->pdata->out_type;
	p->disp_ss_regs = decon->pdata->ss_regs;
}

static void decon_to_psr_info(struct decon_device *decon,
				struct decon_mode_info *psr)
{
	psr->psr_mode = decon->pdata->psr_mode;
	psr->trig_mode = decon->pdata->trig_mode;
	psr->dsi_mode = decon->pdata->dsi_mode;
	psr->out_type = decon->pdata->out_type;

}

/* ---------- FB_IOCTL INTERFACE ----------- */
static int decon_wait_for_vsync(struct decon_device *decon, u32 timeout)
{
	return decon_reg_wait_for_update_timeout(decon->id, timeout * 1000);
}

static void decon_parse_pdata(struct decon_device *decon, u32 dev_id)
{
	struct exynos_decon_platdata *pdata = get_decon_pdata();

	decon->id = dev_id;
	decon->pdata->ip_ver = pdata->ip_ver;
	decon->pdata->psr_mode = pdata->psr_mode;
	decon->pdata->trig_mode = pdata->trig_mode;
	decon->pdata->dsi_mode = pdata->dsi_mode;
	decon->pdata->max_win = pdata->max_win;
	decon->pdata->ss_regs = pdata->ss_regs;
	decon->pdata->default_win = pdata->default_win;

	decon_info("decon-%s: ver%d, max win%d, %s mode, %s trigger\n",
			decon->id ? "ext" : "int", decon->pdata->ip_ver,
			decon->pdata->max_win,
			decon->pdata->psr_mode ? "command" : "video",
			decon->pdata->trig_mode ? "sw" : "hw");
	decon_info("%s dsi mode\n", decon->pdata->dsi_mode ? "dual" : "single");
}

#if defined(CONFIG_DISPLAY_LOGO) || defined(CONFIG_DISPLAY_DRAWFONT)
void decon_show_buffer_update(struct decon_device *decon,
		struct dsim_device *dsim, u32 color)
{
	struct decon_mode_info psr = {0};

	decon_to_psr_info(decon, &psr);
	decon_reg_start(decon->id, &psr);

	decon_wait_for_vsync(decon, VSYNC_TIMEOUT_MSEC);
	if (decon_reg_wait_update_done_and_mask(decon->id, &psr, SHADOW_UPDATE_TIMEOUT) < 0)
		decon_err("%s: wait_for_update_timeout\n", __func__);
}

static void decon_show_buffer(struct decon_device *decon,
		struct dsim_device *dsim, u32 color)
{
	struct decon_window_regs win_regs = {0};
	struct decon_mode_info psr = {0};
	struct decon_param p = {0};
	unsigned int *buf_dup;
	u32 logo_l, logo_t, logo_r, logo_b;
	u32 timeout = 1000;

	decon_to_init_param(decon, &p);
	decon_reg_init(decon->id, 0, &p);

#ifndef CONFIG_DISPLAY_DRAWFONT
#ifdef CONFIG_DISPLAY_LOGO_BASE_ADDRESS
	buf_dup = CONFIG_DISPLAY_LOGO_BASE_ADDRESS;
#else
	buf_dup = kzalloc(LOGO_BUF_SZ<<2, GFP_KERNEL);
	if (!buf_dup) {
		decon_err("no memory for decon frame buffer\n");
		return;
	}
#endif
#endif

	/* stop smmu before proceeding for buffer rendering */
	writel(0, SYSMMU_DISP0_1_BASE_ADDR);

	win_regs.wincon |= WIN_CONTROL_ALPHA_SEL_F(W_ALPHA_SEL_F_BYAEN) | WIN_CONTROL_FUNC_F(PD_FUNC_COPY) \
		| WIN_CONTROL_ALPHA0_F(0xFF) | WIN_CONTROL_ALPHA1_F(0xFF);
	win_regs.wincon |= WIN_CONTROL_EN_F;
	win_regs.start_pos = WIN_STRPTR_Y_F(0) | WIN_STRPTR_X_F(0);
	win_regs.end_pos = (WIN_ENDPTR_Y_F(0 + p.lcd_info->yres - 1) \
		| WIN_ENDPTR_X_F(0 + p.lcd_info->xres - 1));

	win_regs.colormap = 0x0;
	win_regs.pixel_count = p.lcd_info->xres * p.lcd_info->yres;
	win_regs.whole_w = p.lcd_info->xres;
	win_regs.whole_h = p.lcd_info->yres;
	win_regs.offset_x = 0;
	win_regs.offset_y = 0;
	win_regs.type = IDMA_G1;
	decon_reg_set_window_control(decon->id, 0, &win_regs, false);

	decon_reg_update_req_window(decon->id, 0);

	decon_to_psr_info(decon, &psr);
	decon_reg_set_int(decon->id, &psr, 1);
	decon_reg_start(decon->id, &psr);

	call_panel_ops(dsim, displayon, dsim);
	decon_wait_for_vsync(decon, VSYNC_TIMEOUT_MSEC);
	if (decon_reg_wait_update_done_and_mask(decon->id, &psr, SHADOW_UPDATE_TIMEOUT) < 0)
		decon_err("%s: wait_for_update_timeout\n", __func__);

#if !defined(CONFIG_DISPLAY_LOGO_BASE_ADDRESS) && !defined(CONFIG_DISPLAY_DRAWFONT)
	free(buf_dup);
#endif
}

#else
static void decon_show_color_map(struct decon_device *decon,
		struct dsim_device *dsim)
{
	struct decon_window_regs win_regs = {0};
	struct decon_mode_info psr = {0};
	struct decon_param p = {0};

	decon_to_init_param(decon, &p);
	decon_reg_init(decon->id, 0, &p);

	win_regs.wincon |= WIN_CONTROL_ALPHA_SEL_F(W_ALPHA_SEL_F_BYAEN) | WIN_CONTROL_FUNC_F(PD_FUNC_COPY) \
		| WIN_CONTROL_ALPHA0_F(0xFF) | WIN_CONTROL_ALPHA1_F(0xFF);
	win_regs.wincon |= WIN_CONTROL_EN_F;
	win_regs.start_pos = WIN_STRPTR_Y_F(0) | WIN_STRPTR_X_F(0);
	win_regs.end_pos = (WIN_ENDPTR_Y_F(0 + p.lcd_info->yres - 1) \
		| WIN_ENDPTR_X_F(0 + p.lcd_info->xres - 1));

	win_regs.colormap = 0x00FF00;
	win_regs.pixel_count = p.lcd_info->xres * p.lcd_info->yres;
	win_regs.whole_w = p.lcd_info->xres;
	win_regs.whole_h = p.lcd_info->yres;
	win_regs.offset_x = 0;
	win_regs.offset_y = 0;
	win_regs.type = IDMA_G1;
	decon_reg_set_window_control(decon->id, 0, &win_regs, true);

	decon_reg_update_req_window(decon->id, 0);

	decon_to_psr_info(decon, &psr);
	decon_reg_set_int(decon->id, &psr, 1);
	decon_reg_start(decon->id, &psr);

	call_panel_ops(dsim, displayon, dsim);

	decon_wait_for_vsync(decon, VSYNC_TIMEOUT_MSEC);
	if (decon_reg_wait_update_done_and_mask(decon->id, &psr, SHADOW_UPDATE_TIMEOUT) < 0)
		decon_err("%s: wait_for_update_timeout\n", __func__);

}
#endif

/* --------- DRIVER INITIALIZATION ---------- */
static int decon_probe(u32 dev_id)
{
	struct decon_device *decon;
	struct dsim_device *dsim;

	if (dev_id > MAX_DECON_ID) {
		decon_err("does not support (%d) decon device\n", dev_id);
		return -ERANGE;
	}

	decon = kzalloc(sizeof(struct decon_device), GFP_KERNEL);
	if (!decon) {
		decon_err("no memory for decon device\n");
		return -ENOMEM;
	}

	/* setup pointer to master device */
	decon->pdata = kzalloc(sizeof(struct exynos_decon_platdata),
						GFP_KERNEL);
	if (!decon->pdata) {
		decon_err("no memory for DECON platdata\n");
		return -ENOMEM;
	}

	/* store platform data ptr to decon_tv context */
	decon_parse_pdata(decon, dev_id);

	/* init clock setting for decon */
	if (!decon->id) {
		decon_int_drvdata = decon;
		dsim = dsim0_for_decon;
		decon->regs = DECON_INT_BASE_ADDR;
		decon->default_idma = IDMA_G1;
	} else {
		decon_ext_drvdata = decon;
		dsim = dsim1_for_decon;
		decon->regs = DECON_EXT_BASE_ADDR;
	}

	if (!decon->id) {
#if defined(CONFIG_DISPLAY_LOGO) || defined(CONFIG_DISPLAY_DRAWFONT)
		vpp_set_config(IDMAG1_ID, win_fb0);
		decon_show_buffer(decon, dsim, LUT_COLOR_1);
#else
		decon_show_color_map(decon, dsim);
#endif
	}

	decon_info("decon%d registered successfully\n", decon->id);

	return 0;
}

#ifdef CONFIG_DISPLAY_LOGO
void decon_draw_logo(u32 color)
{
	decon_show_buffer(decon_int_drvdata, dsim0_for_decon, color);
}
#endif

void decon_string_update(void)
{
	decon_show_buffer_update(decon_int_drvdata, dsim0_for_decon, LUT_COLOR_1);
}

/* --------- DRIVER DISPLAY START ---------- */
int drv_video_init(void)
{
	struct exynos_decon_platdata *pdata = get_decon_pdata();
	int ret = 0;
	int i;

	for (i = 0; i <= pdata->dsi_mode; i++) {
		ret = dsim_probe(i);
		if (ret < 0) {
			decon_err("dsim%d probe was failed\n", i);
			return ret;
		}
	}

	ret = decon_probe(0);
	if (ret < 0) {
		decon_err("decon%d probe was failed\n", i);
		return ret;
	}

	return ret;
}
